import io
import math
import os.path
import re
from dataclasses import dataclass
from pathlib import Path
from zipfile import ZipFile, is_zipfile

from fs import open_fs

from plateaukit import utils
from plateaukit.readers.citygml import PLATEAUCityGMLParser

# from plateaukit.utils import dict_key_to_camel_case
from .ir_models import IRDocument, IRMetadata


class Readable:
    def __init__(self, files, *args, reader_cls, transformers=[], **kwargs):
        self.reader_cls = reader_cls
        self.reader = reader_cls()
        self.files = files
        self.transformers = transformers
        self.args = args
        self.kwargs = kwargs

    def read(self):
        # TODO: Refactoring
        document = self.reader.read_files(self.files, *self.args, **self.kwargs)

        for transformer in self.transformers:
            document = transformer.transform_document(document)

        return document

    def split(self, split_n):
        group_size = math.ceil(len(self.files) / split_n)

        return [
            Readable(
                files,
                *self.args,
                reader_cls=self.reader_cls,
                **self.kwargs,
            )
            for files in utils.chunker(self.files, group_size)
        ]

    def __len__(self):
        return len(self.files)


@dataclass
class ConcurrentStatus:
    quit = None


# TODO: Derive PLATEAUCityGMLReader or ZippedCityGMLReader
class CityGMLReader:
    def __init__(self, concurrent_status=ConcurrentStatus()):
        self.concurrent_status = concurrent_status

    def _make_codelist_file_map(self, codelist_infiles, base_path, *, zip_fs=None):
        _open = zip_fs.open if zip_fs else open

        codelist_file_map = dict()

        for codelist_infile in codelist_infiles:
            with _open(codelist_infile, "rb") as f:
                relative_path = os.path.relpath(codelist_infile, base_path)
                codelist_file_map[relative_path] = io.BytesIO(f.read())

        return codelist_file_map

    def scan_files(
        self,
        infiles,
        *,
        # object_types,
        # lod,
        codelist_infiles=None,
        zipfile=None,
        selection: list[str] | None = None,
    ):
        """Return a Readable object for the given files."""

        if not codelist_infiles:
            # Codelists
            codelist_infiles = []
            pat = re.compile(r".*codelists\/.*\.xml$")

            if zipfile and is_zipfile(zipfile):
                with ZipFile(zipfile) as f:
                    namelist = f.namelist()
                    codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
                    # NOTE: zipfs requires POSIX path
                    codelist_infiles = [
                        str(Path("/", target)) for target in codelist_infiles
                    ]
            else:
                # TODO: Test support for non-zip codelists
                raise NotImplementedError()
                # codelist_infiles += [str(Path(file_path, "codelists", "*.xml"))]

        return Readable(
            infiles,
            reader_cls=CityGMLReader,
            codelist_infiles=codelist_infiles,
            zipfile=zipfile,
            selection=selection,
        )

    def read_files(
        self,
        infiles,
        *,
        # object_types,
        # lod,
        codelist_infiles=None,
        zipfile=None,
        target_epsg=4326,  # NOTE: Default to WGS84
        selection: list[str] | None = None,
    ):
        if zipfile is not None:
            zip_fs = open_fs(f"zip://{zipfile}")
        else:
            zip_fs = None

        if not codelist_infiles:
            # Codelists
            codelist_infiles = []
            pat = re.compile(r".*codelists\/.*\.xml$")

            if zipfile and is_zipfile(zipfile):
                with ZipFile(zipfile) as f:
                    namelist = f.namelist()
                    codelist_infiles = list(filter(lambda x: pat.match(x), namelist))
                    # NOTE: zipfs requires POSIX path
                    codelist_infiles = [
                        str(Path("/", target)) for target in codelist_infiles
                    ]
            else:
                # TODO: Test support for non-zip codelists
                raise NotImplementedError()
                # codelist_infiles += [str(Path(file_path, "codelists", "*.xml"))]

        # Load codelists
        codelist_file_map = {}
        if infiles:
            base_path = Path(infiles[0]).parent  # TODO: Fix this
            codelist_file_map = self._make_codelist_file_map(
                codelist_infiles, base_path, zip_fs=zip_fs
            )

        parser = PLATEAUCityGMLParser(
            target_epsg=target_epsg, codelist_file_map=codelist_file_map
        )

        _open = zip_fs.open if zip_fs else open

        city_objects = []
        for i, infile in enumerate(infiles):
            with _open(infile, "rb") as f:
                co_iter = parser.iterparse(f, selection=selection)

                for city_obj in co_iter:
                    # if object_types is not None and city_obj.type not in object_types:
                    #     continue
                    if (
                        self.concurrent_status.quit
                        and self.concurrent_status.quit.is_set()
                    ):
                        exit()

                    city_objects.append(city_obj)

        if zip_fs:
            zip_fs.close()

        return IRDocument(
            metadata=IRMetadata(epsg=target_epsg),  # TODO: Fix
            city_objects=city_objects,
        )



import pyproj

from plateaukit.readers.citygml.ir_models import IRDocument, IRGeometry
from plateaukit.readers.citygml.reader import Readable

from .geometry import GeometryTransformer

BOUNDARIES_DEPTH_MAPPING = {
    "MultiPoint": 1,
    "MultiLineString": 2,
    "MultiSurface": 3,
    "CompositeSurface": 3,
    "Solid": 4,
    "MultiSolid": 5,
    "CompositeSolid": 5,
}


class ReprojectionTransformer(GeometryTransformer):
    def __init__(
        self,
        *,
        target_epsg: int,
    ):
        self.target_epsg = target_epsg

    def _reproject_geometry(self, geometry: IRGeometry):
        depth = BOUNDARIES_DEPTH_MAPPING.get(geometry.type, None)

        if depth is None:
            raise ValueError(f"Unsupported geometry type: {geometry.type}")

        # TODO: Type checking
        if depth == 1:
            new_bounds = list(self.transformer.itransform(geometry.boundaries))
        elif depth == 2:
            new_bounds = [
                list(self.transformer.itransform(bounds))
                for bounds in geometry.boundaries
            ]
        elif depth == 3:
            new_bounds = [
                [list(self.transformer.itransform(bounds)) for bounds in surface]
                for surface in geometry.boundaries
            ]
        elif depth == 4:
            new_bounds = [
                [
                    [list(self.transformer.itransform(bounds)) for bounds in region]
                    for region in surface
                ]
                for surface in geometry.boundaries
            ]
        elif depth == 5:
            new_bounds = [
                [
                    [
                        [list(self.transformer.itransform(bounds)) for bounds in solid]
                        for solid in composite
                    ]
                    for composite in geometry.boundaries
                ]
            ]
        else:
            raise ValueError(f"Invalid depth: {depth}")

        geometry.boundaries = new_bounds

        return geometry

    def transform(self, readable: Readable):
        readable.transformers.append(self)

        return readable

    def transform_document(self, document: IRDocument):
        origin_epsg = document.metadata.epsg  # TODO: Check this

        self.transformer = pyproj.Transformer.from_crs(origin_epsg, self.target_epsg)

        for city_object in document.city_objects:
            for i, geom in enumerate(city_object.geometry):
                city_object.geometry[i] = self._reproject_geometry(geom)

        document.metadata.epsg = self.target_epsg

        return document



from dataclasses import dataclass
from typing import Any


@dataclass
class IRGeometry:
    """A CityGML Geometry."""

    type: str
    lod: str
    boundaries: list[Any]
    semantics: dict[str, Any] | None = None


@dataclass
class IRCityObject:
    """A CityGML CityObject."""

    type: str
    id: str | None  # TODO: Make id required
    geometry: list[IRGeometry]
    attributes: dict[str, Any] | None = None
    # children: list[Self] | None = None
    # parent: Self | None = None


@dataclass
class IRMetadata:
    epsg: int


@dataclass
class IRDocument:
    """A CityGML document (Metadata and a collection of CityObjects)"""

    metadata: Any
    city_objects: list[IRCityObject]



from os import PathLike
from typing import Literal, Sequence

import geopandas as gpd

from plateaukit.core.dataset._area import (
    area_from_address,
    area_from_bbox,
    area_from_landmark,
    area_from_points,
    area_from_polygons,
    area_from_postcode,
    get_area,
)
from plateaukit.core.dataset._to_cityjson import to_cityjson
from plateaukit.core.dataset._to_geojson import to_geojson

try:
    from pyogrio import read_dataframe
except ImportError:
    read_dataframe = None

from plateaukit.config import Config
from plateaukit.logger import logger


class Dataset:
    """The Dataset class.

    Attributes:
        dataset_id: Dataset ID.
        gdf: The GeoDataFrame of the dataset.
        object_types: CityGML object types to include in the dataset.
    """

    dataset_id: str
    gdf: gpd.GeoDataFrame | None
    object_types: list[str] | None
    _gdf_cache: dict[str, gpd.GeoDataFrame] | None

    def __init__(self, dataset_id: str, *, object_types: list[str] | None = None):
        """Initialize a dataset.

        Args:
            dataset_id: Dataset ID
            object_types: CityGML object types to include in the dataset.
        """

        self.dataset_id = dataset_id
        self.gdf = None
        self.object_types = object_types

        self._gdf_cache = None

        # config = Config()
        # gpkg_path = config.data[dataset_id].get("gpkg")
        # self.gdf = read_dataframe(gpkg_path)

    def __repr__(self):
        return f"Dataset({self.dataset_id})"

    def load_gdf(self, format: Literal["parquet", "gpkg"] = "parquet"):
        """Load the GeoDataFrame from the prebuilt dataset."""

        if self._gdf_cache is not None:
            return self._gdf_cache

        self._gdf_cache = {}

        # NOTE: Fallback to GeoPackage
        def load_gdf_gpkg():
            if self._gdf_cache is None:
                self._gdf_cache = {}

            gpkg_path = config.datasets[self.dataset_id].get("gpkg")

            if gpkg_path:
                if not read_dataframe:
                    raise ImportError(
                        "Package `pyogrio` is required. Please install it using `pip install pyogrio`."
                    ) from None
                self._gdf_cache["bldg"] = read_dataframe(gpkg_path)  # TODO: Fix typing
            else:
                raise RuntimeError(
                    "Missing a prebuilt dataset; Please run `plateaukit prebuild <dataset_id>` first"
                )

            return self._gdf_cache

        config = Config()

        if format == "parquet":
            logger.debug("Loading Parquet file...")
            parquet_path_or_pathmap = config.datasets[self.dataset_id].get("parquet")

            if isinstance(
                parquet_path_or_pathmap, str
            ):  # NOTE For backward compatibility
                parquet_pathmap = {
                    "bldg": parquet_path_or_pathmap,
                }
            else:
                parquet_pathmap = parquet_path_or_pathmap

            if parquet_pathmap is not None:
                # parquet_path = parquet_paths[0]  # TODO: Fix this
                for k, parquet_path in parquet_pathmap.items():
                    if self.object_types and k not in self.object_types:
                        continue

                    self._gdf_cache[k] = gpd.read_parquet(parquet_path)

                # print(self._gdf_cache)
                return self._gdf_cache
            else:
                return load_gdf_gpkg()

        elif format == "gpkg":
            return load_gdf_gpkg()

        else:
            raise ValueError("Invalid format")

    def get_area(self):
        """Get the entire area of the dataset."""

        return get_area(self)

    def area_from_bbox(self, bbox: list[float] | None = None):
        """Get the specified area from the dataset.

        Args:
            bbox: Bounding box of the area of interest.
                   If not specified, the area of the entire dataset will be returned.
        """

        return area_from_bbox(self, bbox)

    def area_from_polygons(self, polygons: list[gpd.GeoDataFrame]):
        """Get an area from the dataset by polygons.

        Args:
            polygon: Polygon of the area of interest.
        """

        return area_from_polygons(self, polygons)

    def area_from_points(
        self, points: list[Sequence[float]], size: list[float] = [1000, 1000]
    ):
        """Get an area from the dataset by points.

        Args:
            points: List of points of the area of interest.
            size: Size of the area of interest in meters.
        """

        return area_from_points(self, points, size=size)

    def area_from_address(self, address: str, size: list[float] = [1000, 1000]):
        """Get an area from the dataset by an address.

        Args:
            address: Address of the area of interest.
            size: Size of the area of interest in meters.
        """

        return area_from_address(self, address, size=size)

    def area_from_postcode(self, postcode: str):
        """Get an area from the dataset by a postcode.

        Args:
            postcode: Postal code of the area of interest.
        """

        return area_from_postcode(self, postcode)

    def area_from_landmark(self, landmark: str, min_size: list[float] = [1000, 1000]):
        """Get an area from the dataset by a landmark.

        Args:
            landmark: Landmark of the area of interest.
        """

        return area_from_landmark(self, landmark, min_size=min_size)

    # def search_area(self, keyword: str):
    #     print(keyword)

    def to_geojson(
        self,
        outfile: str | PathLike,
        *,
        types: list[str] = ["bldg"],
        altitude: bool = True,
        include_type: bool = False,
        seq=False,
        split: int = 1,
        **kwargs,
    ):
        """Export GeoJSON from PLATEAU datasets.

        Args:
            outfile: Output file path.
            types: CityGML object types to include.
            split: Split the output into specified number of files.
            **kwargs: Keyword arguments for the generator.
        """

        return to_geojson(
            self,
            outfile,
            types=types,
            altitude=altitude,
            include_type=include_type,
            seq=seq,
            split=split,
            **kwargs,
        )

    def to_cityjson(
        self,
        outfile: str | PathLike,
        *,
        types: list[str] = ["bldg"],
        object_types=None,  # TODO: Handle this
        lod_mode: Literal["highest", "all", "values"] = "highest",
        lod_values: list[str] | None = None,
        ground: bool = False,
        seq: bool = False,
        split: int = 1,
        selection: list[str] | None = None,
        target_epsg: int | None = None,
        progress_messages: dict | None = None,
        **kwargs,
    ):
        """Export CityJSON from PLATEAU datasets.

        Args:
            outfile: Output file path.
            types: CityGML feature types.
            split: Split the output into specified number of files.
            seq: Export CityJSONSeq.
            **kwargs: Keyword arguments for the generator.
        """

        return to_cityjson(
            self,
            outfile,
            types=types,
            object_types=object_types,
            lod_mode=lod_mode,
            lod_values=lod_values,
            ground=ground,
            seq=seq,
            split=split,
            selection=selection,
            target_epsg=target_epsg,
            progress_messages=progress_messages,
            **kwargs,
        )


def load_dataset(dataset_id: str, object_types: list[str] | None = None) -> Dataset:
    """Load a dataset.

    Args:
        dataset_id: Dataset ID.
        object_types: CityGML object types to include in the dataset. If not specified, all object types will be included.

    Returns:
        Dataset: Dataset object.
    """
    # TODO: Raise error if dataset_id is not in local datasets and not a cloud dataset

    return Dataset(dataset_id, object_types=object_types)



import sys
from typing import Sequence, cast

import flatgeobuf as fgb
import geopandas as gpd

try:
    from pyogrio import read_dataframe
except ImportError:
    read_dataframe = None

from plateaukit import defaults, geocoding
from plateaukit.core.area import Area, GeoDataFrameLayer


def get_area(self):
    """Get the entire area of the dataset."""

    if self.dataset_id.endswith(".cloud"):
        raise Exception("Not supported for cloud datasets")

    dataframes = self.load_gdf()

    layers = {}

    for k, gdf in dataframes.items():
        if self.object_types is None or k in self.object_types:
            layers[k] = GeoDataFrameLayer(gdf)

    if "bldg" not in layers:
        raise RuntimeError("Missing building data")

    area = Area(layers, base_layer_name="bldg")
    area._datasets = [self.dataset_id]

    return area


def area_from_bbox(self, bbox: list[float] | None = None):
    """Get the specified area from the dataset.

    Args:
        bbox: Bounding box of the area of interest.
                If not specified, the area of the entire dataset will be returned.
    """

    if self.dataset_id.endswith(".cloud"):
        if "pyodide" in sys.modules:
            try:
                import pyodide_http  # type: ignore

                pyodide_http.patch_all()
            except Exception:
                pass

        # Load data from remote flatgeobuf
        remote_fgb = defaults.cloud_base_url + self.dataset_id.replace(".cloud", ".fgb")
        if read_dataframe:
            area_gdf = read_dataframe(
                remote_fgb,
                bbox=tuple(bbox) if bbox else None,
            )
            area_gdf = cast(gpd.GeoDataFrame, area_gdf)
        else:
            feature_collection = fgb.load_http(
                remote_fgb,
                bbox=tuple(bbox) if bbox else None,
            )
            area_gdf = gpd.GeoDataFrame.from_features(feature_collection)

        layers = {}
        layers["bldg"] = GeoDataFrameLayer(area_gdf)
        area = Area(layers, base_layer_name="bldg")

    else:
        dataframes = self.load_gdf()

        layers = {}

        for k, gdf in dataframes.items():
            if self.object_types is None or k in self.object_types:
                # TODO: Error handling when area_gdf is empty
                area_gdf = gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]] if bbox else gdf
                layers[k] = GeoDataFrameLayer(area_gdf)

        if "bldg" not in layers:
            raise RuntimeError("`bldg` layer is required")

        area = Area(layers, base_layer_name="bldg")

    area._datasets = [self.dataset_id]

    return area


def area_from_polygons(self, polygons: list[gpd.GeoDataFrame]):
    """Get an area from the dataset by polygons.

    Args:
        polygon: Polygon of the area of interest.
    """

    if self.dataset_id.endswith(".cloud"):
        raise Exception("Not supported for cloud datasets")

    dataframes = self.load_gdf()

    layers = {}

    for k, gdf in dataframes.items():
        if self.object_types is None or k in self.object_types:
            # TODO: Error handling when area_gdf is empty
            area_gdf = gdf[gdf.geometry.intersects(polygons)]
            layers[k] = GeoDataFrameLayer(area_gdf)

    if "bldg" not in layers:
        raise RuntimeError("`bldg` layer is required")

    area = Area(layers, base_layer_name="bldg")

    return area


def area_from_points(
    self, points: list[Sequence[float]], size: list[float] = [1000, 1000]
):
    """Get an area from the dataset by points.

    Args:
        points: List of points of the area of interest.
        size: Size of the area of interest in meters.
    """

    bbox = geocoding._get_bbox(points)
    bbox = geocoding._pad_bbox(bbox, size)

    return self.area_from_bbox(bbox)


def area_from_address(self, address: str, size: list[float] = [1000, 1000]):
    """Get an area from the dataset by an address.

    Args:
        address: Address of the area of interest.
        size: Size of the area of interest in meters.
    """

    bbox = geocoding.bbox_from_address(address, min_size=size)

    return self.area_from_bbox(bbox)


def area_from_postcode(self, postcode: str):
    """Get an area from the dataset by a postcode.

    Args:
        postcode: Postal code of the area of interest.
    """

    bbox = geocoding.bbox_from_postcode(postcode)

    return self.area_from_bbox(bbox)


def area_from_landmark(self, landmark: str, min_size: list[float] = [1000, 1000]):
    """Get an area from the dataset by a landmark.

    Args:
        landmark: Landmark of the area of interest.
    """
    bbox = geocoding.bbox_from_landmark(landmark, min_size=min_size)

    return self.area_from_bbox(bbox)



from abc import ABC, abstractmethod

from geopandas import GeoDataFrame


class BaseLayer(ABC):
    """The BaseLayer class."""

    name: str | None

    @abstractmethod
    def get_area(self, bbox: list[float] | None = None) -> "BaseLayer":
        pass


class GeoDataFrameLayer(BaseLayer):
    """The GeoDataFrameLayer class."""

    gdf: GeoDataFrame

    def __init__(self, gdf: GeoDataFrame, *, name: str | None = None) -> None:
        self.gdf = gdf
        self.name = name

    def __repr__(self) -> str:
        return "GeoDataFrameLayer()"

    def get_area(self, bbox: list[float] | None = None):
        area_gdf = (
            self.gdf.cx[bbox[0] : bbox[2], bbox[1] : bbox[3]] if bbox else self.gdf
        )

        # TODO: Error handling when area_gdf is empty

        return GeoDataFrameLayer(area_gdf)



from plateaukit.readers.citygml.ir_models import IRDocument

from .geometry import GeometryTransformer


class ShiftToGroundTransformer(GeometryTransformer):
    def _get_nesting_level(self, boundaries: list | None):
        if isinstance(boundaries, list):
            return 1 + max([self._get_nesting_level(x) for x in boundaries])
        else:
            return 0

    def _get_min_z(self, boundaries: list):
        level = self._get_nesting_level(boundaries)

        # Nest level 3
        if level == 3:
            min_z = min(
                [
                    min([min([point[2] for point in region]) for region in surface])
                    for surface in boundaries
                ]
            )
        # Nest level 4
        elif level == 4:
            min_z = min(
                [
                    min(
                        [
                            min([min([point[2] for point in shell]) for shell in solid])
                            for solid in surface
                        ]
                    )
                    for surface in boundaries
                ]
            )
        else:
            raise NotImplementedError()

        return min_z

    def _shift_boundaries_to_ground(self, boundaries: list, min_z: float):
        level = self._get_nesting_level(boundaries)

        # Nest level 3
        if level == 3:
            boundaries = [
                [
                    [(point[0], point[1], point[2] - min_z) for point in region]
                    for region in surface
                ]
                for surface in boundaries
            ]
        # Nest level 4
        elif level == 4:
            boundaries = [
                [
                    [
                        [(point[0], point[1], point[2] - min_z) for point in shell]
                        for shell in solid
                    ]
                    for solid in surface
                ]
                for surface in boundaries
            ]

        return boundaries

    def transform_document(self, document: IRDocument):
        for city_object in document.city_objects:
            min_z = 0xFFFF

            for geometry in city_object.geometry:
                min_z = min(min_z, self._get_min_z(geometry.boundaries))
                geometry.boundaries = self._shift_boundaries_to_ground(
                    geometry.boundaries, min_z
                )

        return document



from textwrap import dedent
from typing import Literal

from plateaukit.core.widgets.interactive_deck import InteractiveDeck


def _is_colab():
    import sys

    return "google.colab.output" in sys.modules


def set_llm(self, llm):
    self.llm = llm


def chat(
    self, message: str, *, output_format: Literal["map", "dataframe"] | None = "map"
):
    """LLM-based chat interface for the area of interest using LangChain."""

    import geopandas as gpd

    try:
        from langchain_core.output_parsers.openai_tools import (
            JsonOutputKeyToolsParser,
        )
        from langchain_core.prompts import ChatPromptTemplate
        from langchain_experimental.tools import PythonAstREPLTool
    except ImportError:
        raise ImportError(
            "Please install langchain-core and langchain-experimental to use this feature."
        )

    if not hasattr(self, "llm") or self.llm is None:
        raise RuntimeError("Set LangChain's LLM instance with set_llm() method first.")

    llm = self.llm

    df = self.gdf.copy()

    tool = PythonAstREPLTool(globals={"df": df}, verbose=True)
    tool.locals = tool.globals
    llm_with_tools = llm.bind_tools([tool], tool_choice=tool.name)
    parser = JsonOutputKeyToolsParser(key_name=tool.name, first_tool_only=True)

    system = dedent(
        f"""You have access to a geopandas dataframe called `df`.
        Please output Python code in response to the user's instructions. Do not output anything else.
        The following is the result of executing `df.head().to_markdown()`:
        {df.head().to_markdown()}
        The following is a description of the items in this dataframe:
        - buildingId: The ID of the building
        - name: The name of the building. Always look for the name of the building in this field
        - measuredHeight: The measured height of the building
        - storeysAboveGround: The number of floors above ground
        - storeysBelowGround: The number of floors below ground
        - usage: The usage of the building
        - longitude: The latitude of the building
        - latitude: The longitude of the building
        The following is a list of buildings whose names exist in this dataframe:
        {df['name'].unique()}
        The following is a list of the usage types of buildings that exist in this data frame:
        {df['usage'].unique()}
        Again, please output Python code in response to the user's instructions. Do not output anything else.
        Do not use any knowledge about the coordinates other than the dataframe.
        Only the standard Python libraries, pandas, and geopandas can be used as libraries.
        Be sure to write Python code that returns a pandas dataframe."""
    )

    prompt = ChatPromptTemplate.from_messages(
        [("system", system), ("human", "{input}")]
    )

    chain = prompt | llm_with_tools | parser | tool

    response = chain.invoke(message)

    if isinstance(response, gpd.GeoDataFrame) and output_format == "map":
        deck = InteractiveDeck(response)

        self.selection = deck.selection

        if _is_colab():
            return deck.deck
        else:
            return deck.widget

    return response



from plateaukit.readers.citygml.parsers.xml_models.city_object import CityObjectXML


def _get_string_attribute(xml: CityObjectXML, name) -> str | None:
    path = f"./gen:stringAttribute[@name='{name}']/gen:value"
    result = xml.find(path, xml.nsmap)
    return result.text if result is not None else None


def get_name(xml: CityObjectXML) -> str | None:
    value = xml._get_codespace_attribute("./gml:name")
    return value


def get_usage(xml: CityObjectXML) -> str | None:
    value = xml._get_codespace_attribute("./bldg:usage")
    return value


def get_district_plan(xml: CityObjectXML) -> str | None:
    value = _get_string_attribute(xml, name="地区計画")
    return value


def get_building_id(xml: CityObjectXML) -> str | None:
    try:
        path = "./uro:buildingIDAttribute/uro:BuildingIDAttribute/uro:buildingID"
        result = xml.find(path, xml.nsmap)
        value = result.text if result is not None else None
        assert value is not None
        return value
    except AssertionError:
        pass

    try:
        value = _get_string_attribute(xml, name="建物ID")
        assert value is not None
        return value
    except AssertionError:
        pass

    return None


def get_measured_height(xml: CityObjectXML) -> float | None:
    result = xml.find("./bldg:measuredHeight", xml.nsmap)
    value = result.text if result is not None else None
    value = float(value) if value is not None else None
    return value


def get_year_of_construction(xml: CityObjectXML) -> int | None:
    result = xml.find("./bldg:yearOfConstruction", xml.nsmap)
    value = result.text if result is not None else None
    value = int(value) if value is not None else None
    return value


def get_storeys_above_ground(xml: CityObjectXML) -> int | None:
    result = xml.find("./bldg:storeysAboveGround", xml.nsmap)
    value = result.text if result is not None else None
    value = int(value) if value is not None else None
    return value


def get_storeys_below_ground(xml: CityObjectXML) -> int | None:
    result = xml.find("./bldg:storeysBelowGround", xml.nsmap)
    value = result.text if result is not None else None
    value = int(value) if value is not None else None
    return value


def get_river_flooding_risks(xml: CityObjectXML):
    results = xml.iterfind(
        "./uro:buildingDisasterRiskAttribute/uro:BuildingRiverFloodingRiskAttribute",
        xml.nsmap,
    )
    value = {}
    for result in results:
        description = xml._get_codespace_attribute("./uro:description", parent=result)

        # rank = xml._get_codespace_attribute("./uro:rank", parent=result)
        # rank_org = xml._get_codespace_attribute("./uro:rankOrg", parent=result)

        depth = result.find("./uro:depth", xml.nsmap)
        depth = float(depth.text) if depth is not None else None

        duration = result.find("./uro:duration", xml.nsmap)
        duration = float(duration.text) if duration is not None else None

        admin_type = xml._get_codespace_attribute("./uro:adminType", parent=result)

        scale = xml._get_codespace_attribute("./uro:scale", parent=result)

        value[description] = {
            # "rank": rank,
            # "rankOrg": rank_org,
            "depth": depth,
            "duration": duration,
            "adminType": admin_type,
            "scale": scale,
        }

    return value


def get_river_flooding_depth(xml: CityObjectXML) -> float | None:
    result = xml.find(
        "./uro:buildingDisasterRiskAttribute/uro:BuildingRiverFloodingRiskAttribute/uro:depth",
        xml.nsmap,
    )
    value = result.text if result is not None else None
    value = float(value) if value is not None else None
    return value


def get_river_flooding_duration(xml: CityObjectXML) -> float | None:
    result = xml.find(
        "./uro:buildingDisasterRiskAttribute/uro:BuildingRiverFloodingRiskAttribute/uro:duration",
        xml.nsmap,
    )
    value = result.text if result is not None else None
    value = float(value) if value is not None else None
    return value


def get_districts_and_zones_type(xml: CityObjectXML) -> str | None:
    el = xml.find(
        "./uro:buildingDetailAttribute",
        xml.nsmap,
    )

    if el is None:
        return None

    # TODO:
    # ".//uro:urbanPlanType", # "都市計画区域" など
    # ".//uro:areaClassificationType",  # "市街化区域" など

    # 都市計画法第8条第3項第1号に定める地域地区（及び用途地域）の区分
    result = xml._get_codespace_attribute(
        "./uro:BuildingDetailAttribute/uro:districtsAndZonesType", parent=el
    )

    return result


def get_address(xml: CityObjectXML) -> dict | None:
    el = xml.find("./bldg:address", xml.nsmap)

    if el is None:
        return None

    result = el.find(".//xAL:LocalityName", xml.nsmap)

    if result is None:
        return None

    locality_name = result.text

    addr = {
        "locality_name": locality_name,
    }

    return addr



from __future__ import annotations

from plateaukit.cli.base import cli
from plateaukit.cli.commands.config import config_cmd
from plateaukit.cli.commands.export import (
    export_cityjson_cmd,
    export_geojson_cmd,
    export_qmesh_cmd,
)
from plateaukit.cli.commands.info import info_cmd
from plateaukit.cli.commands.install import install_cmd, list_cmd, uninstall_cmd
from plateaukit.cli.commands.prebuild import prebuild_cmd

cli.add_command(list_cmd)
cli.add_command(install_cmd)
cli.add_command(uninstall_cmd)
cli.add_command(prebuild_cmd)
cli.add_command(export_geojson_cmd)
cli.add_command(export_cityjson_cmd)
cli.add_command(export_qmesh_cmd)
cli.add_command(info_cmd)
cli.add_command(config_cmd)

if __name__ == "__main__":
    cli()



import re
import zipfile
from os import PathLike
from pathlib import Path, PurePosixPath

from plateaukit.config import Config
from plateaukit.logger import logger


def to_geojson(
    self,
    outfile: str | PathLike,
    *,
    types: list[str] = ["bldg"],
    altitude: bool = True,
    include_type: bool = False,
    seq=False,
    split: int = 1,
    **kwargs,
):
    """Export GeoJSON from PLATEAU datasets.

    Args:
        outfile: Output file path.
        types: CityGML object types to include.
        split: Split the output into specified number of files.
        **kwargs: Keyword arguments for the generator.
    """
    # NOTE: exporters requires multiprocessing at the moment, unavailable in pyodide
    from plateaukit import exporters

    if not self.dataset_id:
        raise Exception("Missing dataset_id")

    if not types:
        raise Exception("Missing object types")

    # NOTE: this is intentional but to be refactored in the future
    config = Config()
    record = config.datasets[self.dataset_id]

    if "citygml" not in record:
        raise Exception("Missing CityGML data")
    file_path = Path(record["citygml"])

    infiles = []

    # TODO: Refactor
    for type in types:
        # TODO: Fix
        pat = re.compile(rf".*udx\/{type}\/.*\.gml$")
        if zipfile.is_zipfile(file_path):
            with zipfile.ZipFile(file_path) as f:
                namelist = f.namelist()
                targets = list(filter(lambda x: pat.match(x), namelist))
                # print(targets)

                if not targets:
                    raise RuntimeError(
                        f"Data type '{type}' not found in '{self.dataset_id}'"
                    )

                # NOTE: zipfs requires POSIX path
                infiles += [str(PurePosixPath("/", target)) for target in targets]

                # f.extractall(tdir, members=targets)
                # infiles += [
                #     str(Path(tdir, Path(file_path).stem, "udx", type, "*.gml"))
                # ]
        else:
            infiles += [str(Path(file_path, "udx", type, "*.gml"))]

    logger.debug([types, infiles, outfile])

    # Codelists
    codelist_infiles = None
    if not zipfile.is_zipfile(file_path):
        # TODO: Test support for non-zip codelists
        codelist_infiles = [str(Path(file_path, "codelists", "*.xml"))]

    exporters.geojson.geojson_from_citygml(
        infiles,
        outfile,
        types=types,
        altitude=altitude,
        include_type=include_type,
        split=split,
        seq=seq,
        zipfile=file_path,
        codelist_infiles=codelist_infiles,
        **kwargs,
    )



import glob
import tempfile
from pathlib import Path
from typing import Literal

import geopandas as gpd
import pandas as pd
from rich import get_console

from plateaukit.config import Config
from plateaukit.core.dataset import load_dataset
from plateaukit.logger import logger

from ._cityobjectsdb import _prebuild_cityobjects_db

_supported_types = ["bldg", "brid", "tran"]


def prebuild(
    dataset_id: str,
    *,
    split: int = 10,
    simple_output=False,
    format: Literal["gpkg", "parquet"] = "parquet",
    types=["bldg"],
    prebuild_cityobjects: bool = True,
) -> None:
    """Prebuild a PLATEAU dataset for PlateauKit."""

    try:
        from pyogrio import read_dataframe, write_dataframe
    except ImportError:
        raise ImportError(
            "Package `pyogrio` is required for prebuild. Please install it using `pip install pyogrio`."
        ) from None

    console = get_console()

    if not dataset_id:
        raise Exception("Missing argument: dataset_id")

    config = Config()
    # record = config.datasets.get(dataset_id)
    # print(dataset_id, record)

    # Check if all types are supported
    if not set(types).issubset(set(_supported_types)):
        raise ValueError(f"Unsupported types: {set(types) - set(_supported_types)}")

    with tempfile.TemporaryDirectory() as tdir:
        logger.debug(f"Temporary directory: {tdir}")

        for type in types:
            outfile_geojsonl = Path(tdir, f"{dataset_id}.{type}.geojsonl")

            dataset = load_dataset(dataset_id)
            dataset.to_geojson(
                outfile_geojsonl,
                types=[type],
                altitude=False,  # TODO: Check this
                include_type=True,
                seq=True,
                split=split,
                progress={"description": f"Generating GeoJSONSeq files: {type}"},
                simple_output=simple_output,
            )

        if format == "parquet":
            display_name = "Parquet files"
        elif format == "gpkg":
            display_name = "GeoPackage files"
        else:
            raise ValueError(f"Invalid format: {format}")

        if prebuild_cityobjects:
            dest_path = str(Path(config.data_dir, f"{dataset_id}.cityobjects.parquet"))

            _prebuild_cityobjects_db(
                dataset_id,
                dest_path,
                types=types,
                tmpdir=tdir,
                split=split,
            )

            config.datasets[dataset_id]["cityobjects"] = dest_path

        with console.status(f"Writing {display_name}...") as status:
            if format == "gpkg":
                if types != ["bldg"]:
                    raise NotImplementedError(
                        "GeoPackage mode supports `bldg` type only."
                    )

                df = gpd.GeoDataFrame()

                for filename in glob.glob(str(Path(tdir, "*.geojsonl"))):
                    # Filter by type
                    subdf = read_dataframe(filename)
                    # NOTE: Setting ignore_index True for re-indexing
                    df = pd.concat([df, subdf], ignore_index=True)

                # TODO: Use more accurate CRS
                mercator = df.to_crs(3857)
                centroid_mercator = mercator.centroid
                centroid = centroid_mercator.to_crs(4326)

                df["longitude"] = centroid.x
                df["latitude"] = centroid.y

                dest_path = Path(config.data_dir, f"{dataset_id}.gpkg")
                write_dataframe(df, dest_path, driver="GPKG")

                config.datasets[dataset_id]["gpkg"] = dest_path
                config.save()

            elif format == "parquet":
                dest_path_map = {}

                for type in types:
                    tmp_file_paths = []

                    for filename in glob.glob(str(Path(tdir, "*.geojsonl"))):
                        # Filter by type
                        if f".{type}." not in filename:
                            continue

                        try:
                            df = gpd.read_file(filename)
                        except Exception:
                            raise RuntimeError(f"Failed to read {filename}")
                            # import shutil

                            # # copy dir to /tmp for debugging:
                            # shutil.copytree(tdir, "/tmp/failed")
                            # raise

                        # TODO: Use more accurate CRS
                        mercator = df.to_crs(3857)
                        centroid_mercator = mercator.centroid
                        centroid = centroid_mercator.to_crs(4326)

                        df["longitude"] = centroid.x
                        df["latitude"] = centroid.y

                        tmp_dest_path = str(Path(tdir, f"{filename}.parquet"))
                        df.to_parquet(tmp_dest_path, index=False)

                        tmp_file_paths.append(tmp_dest_path)

                    df = None
                    for filename in tmp_file_paths:
                        subdf = gpd.read_parquet(filename)
                        # NOTE: set ignore_index True for re-indexing
                        df = (
                            pd.concat([df, subdf], ignore_index=True)
                            if df is not None
                            else subdf
                        )

                    if df is None:
                        raise RuntimeError("Data is empty")

                    # NOTE: For backward compatibility
                    if type == "bldg":
                        dest_filename = f"{dataset_id}.parquet"
                    else:
                        dest_filename = f"{dataset_id}.{type}.parquet"

                    dest_path = Path(config.data_dir, dest_filename)
                    df.to_parquet(dest_path, compression="zstd")

                    dest_path_map[type] = dest_path

                config.datasets[dataset_id]["parquet"] = dest_path_map
                config.save()

        console.print(f"Writing {display_name}... [green]Done")

        # click.echo(f"\nCreated: {dest_path}")



import re
import zipfile
from os import PathLike
from pathlib import Path, PurePosixPath
from typing import Literal

from plateaukit.config import Config
from plateaukit.exporters.cityjson.parallel_writer import ParallelWriter
from plateaukit.exporters.cityjson.writer import CityJSONWriter
from plateaukit.logger import logger
from plateaukit.readers.citygml.reader import CityGMLReader
from plateaukit.transformers.filter_lod import LODFilteringTransformer
from plateaukit.transformers.reprojection import ReprojectionTransformer


def to_cityjson(
    self,
    outfile: str | PathLike,
    *,
    types: list[str] = ["bldg"],
    object_types=None,  # TODO: Handle this
    lod_mode: Literal["highest", "all", "values"] = "highest",
    lod_values: list[str] | None = None,
    ground: bool = False,
    seq: bool = False,
    split: int = 1,
    selection: list[str] | None = None,
    target_epsg: int | None = None,
    progress_messages: dict | None = None,
    **kwargs,
):
    """Export CityJSON from PLATEAU datasets.

    Args:
        outfile: Output file path.
        types: CityGML feature types.
        split: Split the output into specified number of files.
        seq: Export CityJSONSeq.
        **kwargs: Keyword arguments for the generator.
    """
    # NOTE: (check) generators requires multiprocessing at the moment, unavailable in pyodide
    # params = {}

    # if precision:
    #     params["precision"] = precision

    if not self.dataset_id:
        raise Exception("Missing dataset_id")

    config = Config()
    record = config.datasets[self.dataset_id]

    if "citygml" not in record:
        raise Exception("Missing CityGML data")

    file_path = Path(record["citygml"])

    infiles = []

    for type in types:
        # TODO: fix
        pat = re.compile(rf".*udx\/{type}\/.*\.gml$")

        if zipfile.is_zipfile(file_path):
            with zipfile.ZipFile(file_path) as f:
                namelist = f.namelist()
                targets = list(filter(lambda x: pat.match(x), namelist))

                if not targets:
                    raise RuntimeError(
                        f"Data type '{type}' not found in '{self.dataset_id}'"
                    )

                # NOTE: zipfs requires POSIX path
                infiles += [str(PurePosixPath("/", target)) for target in targets]
        else:
            infiles += [str(Path(file_path, "udx", type, "*.gml"))]

        logger.debug([types, infiles, outfile])

    # Sort by a part of filename to group by areas; TODO: Update this
    # try:
    #     infiles = sorted(infiles, key=lambda x: Path(x).stem.split("_")[0])
    # except:
    #     infiles = sorted(infiles)
    infiles = sorted(infiles)

    # Codelists
    codelist_infiles = None
    if not zipfile.is_zipfile(file_path):
        # TODO: Test support for non-zip codelists
        codelist_infiles = [str(Path(file_path, "codelists", "*.xml"))]

    reader = CityGMLReader()

    readable = reader.scan_files(
        infiles,
        codelist_infiles=codelist_infiles,
        zipfile=file_path,
        selection=selection,
    )

    # TODO: Fix typing
    transformers: list = [
        LODFilteringTransformer(mode=lod_mode, values=lod_values),
    ]

    if target_epsg:
        transformers.append(ReprojectionTransformer(target_epsg=target_epsg))

    for transformer in transformers:
        readable = transformer.transform(readable)

    parallel_writer = ParallelWriter(CityJSONWriter)
    parallel_writer.transform(
        readable,
        str(outfile),
        seq=seq,
        split=split,
        progress_messages=progress_messages,
    )

    # generators.cityjson.cityjson_from_citygml(
    #     infiles,
    #     outfile,
    #     split=split,
    #     zipfile=file_path,
    #     lod=lod,
    #     use_highest_lod=use_highest_lod,
    #     ground=ground,
    #     seq=seq,
    #     codelist_infiles=codelist_infiles,
    #     selection=selection,
    #     target_epsg=target_epsg,
    #     **kwargs,
    # )
